@@ -5,9 +5,9 @@ import enum
 from typing import Mapping, Optional, Sequence, Union
 
 from .expr import (
+    EnableExpr,
     IdExpr,
     OptBoolExpr,
-    OptEnableExpr,
     OptIdExpr,
     OptIntExpr,
     OptLifeSpanExpr,
@@ -111,7 +111,6 @@ class Strategy(Base):
     matrix: Matrix
     fail_fast: OptBoolExpr
     max_parallel: OptIntExpr
-    cache: Optional[Cache] = field(metadata={"allow_none": True})
 
 
 @dataclass(frozen=True)
@@ -140,7 +139,8 @@ class TaskBase(Base):
     strategy: Optional[Strategy] = field(metadata={"allow_none": True})
 
     # continue_on_error: OptBoolExpr
-    enable: OptEnableExpr
+    enable: EnableExpr = field(metadata={"default_expr": "${{ success() }}"})
+    cache: Optional[Cache] = field(metadata={"allow_none": True})
 
 
 @dataclass(frozen=True)
@@ -161,7 +161,8 @@ class JobActionCall(BaseActionCall):
 
 @dataclass(frozen=True)
 class TaskActionCall(BaseActionCall, TaskBase):
-    pass
+    enable: EnableExpr = field(metadata={"default_expr": "${{ success() }}"})
+    cache: Optional[Cache] = field(metadata={"allow_none": True})
 
 
 @dataclass(frozen=True)
@@ -230,6 +231,7 @@ class ActionKind(enum.Enum):
     LIVE = "live"  # live composite
     BATCH = "batch"  # batch composite
     STATEFUL = "stateful"  # stateful, can be used in batch flow
+    LOCAL = "local"  # runs locally, can be used in batch flow
 
 
 @dataclass(frozen=True)
@@ -280,9 +282,12 @@ class BatchAction(BaseAction):
 class StatefulAction(BaseAction):
     outputs: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})
     cache: Optional[Cache] = field(metadata={"allow_none": True})
-
-    pre: Optional[ExecUnit] = field(metadata={"allow_none": True})
-    pre_if: OptBoolExpr
     main: ExecUnit
     post: Optional[ExecUnit] = field(metadata={"allow_none": True})
-    post_if: OptBoolExpr
+    post_if: EnableExpr = field(metadata={"default_expr": "${{ always() }}"})
+
+
+@dataclass(frozen=True)
+class LocalAction(BaseAction):
+    outputs: Optional[Mapping[str, Output]] = field(metadata={"allow_none": True})
+    cmd: StrExpr
