@@ -10,13 +10,19 @@ import sys
 from neuromation.api import Client, Factory, JobDescription, JobStatus, ResourceNotFound
 from neuromation.cli.formatters import ftable  # TODO: extract into a separate library
 from types import TracebackType
-from typing import AsyncIterator, Iterable, List, Optional, Tuple, Type
+from typing import AbstractSet, AsyncIterator, Iterable, List, Optional, Tuple, Type
 from typing_extensions import AsyncContextManager
 
+from .config_loader import LiveLocalCL
 from .context import ImageCtx, LiveContext, UnknownJob, VolumeCtx
-from .parser import parse_live
-from .types import LocalPath
-from .utils import format_job_status
+from .parser import ConfigDir
+from .utils import (
+    RUNNING_JOB_STATUSES,
+    TERMINATED_JOB_STATUSES,
+    fmt_id,
+    fmt_raw_id,
+    fmt_status,
+)
 
 
 @dataclasses.dataclass(frozen=True)
@@ -29,24 +35,24 @@ class JobInfo:
 
 
 class LiveRunner(AsyncContextManager["LiveRunner"]):
-    def __init__(self, workspace: LocalPath, config_file: LocalPath) -> None:
-        self._workspace = workspace
-        self._config_file = config_file
-        self._flow = parse_live(workspace, config_file)
+    def __init__(self, config_dir: ConfigDir) -> None:
+        self._config_dir = config_dir
+        self._config_loader: Optional[LiveLocalCL] = None
         self._ctx: Optional[LiveContext] = None
         self._client: Optional[Client] = None
 
     async def post_init(self) -> None:
         if self._ctx is not None:
             return
-        self._ctx = await LiveContext.create(
-            self._flow, self._workspace, self._config_file
-        )
+        self._config_loader = LiveLocalCL(self._config_dir)
+        self._ctx = await LiveContext.create(self._config_loader, "live")
         self._client = await Factory().get()
 
     async def close(self) -> None:
         if self._client is not None:
             await self._client.close()
+        if self._config_loader is not None:
+            await self._config_loader.close()
 
     async def __aenter__(self) -> "LiveRunner":
         await self.post_init()
@@ -92,16 +98,16 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                 if suffix is None:
                     if not skip_check:
                         raise click.BadArgumentUsage(
-                            f"Please provide a suffix for multi-job {job_id}"
+                            f"Please provide a suffix for multi-job {fmt_id(job_id)}"
                         )
             else:
                 if suffix is not None:
                     raise click.BadArgumentUsage(
-                        f"Suffix is not allowed for non-multijob {job_id}"
+                        f"Suffix is not allowed for non-multijob {fmt_id(job_id)}"
                     )
             return ctx
         except UnknownJob:
-            click.secho(f"Unknown job {click.style(job_id, bold=True)}", fg="red")
+            click.secho(f"Unknown job {fmt_id(job_id)}", fg="red")
             jobs_str = ",".join(self.ctx.job_ids)
             click.secho(f"Existing jobs: {jobs_str}", dim=True)
             sys.exit(1)
@@ -113,19 +119,19 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         found = False
         if meta.multi and not suffix:
             async for job in self.client.jobs.list(
-                tags=meta.tags,
+                tags=meta_ctx.tags,
                 reverse=True,
             ):
                 found = True
                 yield job
         else:
-            tags = list(meta.tags)
+            tags = list(meta_ctx.tags)
             if meta.multi and suffix:
                 tags.append(f"multi:{suffix}")
             async for job in self.client.jobs.list(
                 tags=tags,
                 reverse=True,
-                limit=100,  # fixme: limit should be 1 but it doesn't work
+                limit=1,
             ):
                 found = True
                 yield job
@@ -162,11 +168,22 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                         JobInfo(real_id, descr.status, descr.id, descr.tags, when)
                     )
         except ResourceNotFound:
-            ret.append(
-                JobInfo(job_id, JobStatus.UNKNOWN, None, meta_ctx.meta.tags, None)
-            )
+            ret.append(JobInfo(job_id, JobStatus.UNKNOWN, None, meta_ctx.tags, None))
         return ret
 
+    async def list_suffixes(self, job_id: str) -> AbstractSet[str]:
+        meta_ctx = await self._ensure_meta(job_id, None, skip_check=True)
+        result = set()
+        try:
+            async for job in self._resolve_jobs(meta_ctx, None):
+                for tag in job.tags:
+                    key, sep, val = tag.partition(":")
+                    if sep and key == "multi":
+                        result.add(val)
+        except ResourceNotFound:
+            pass
+        return result
+
     async def ps(self) -> None:
         """Return statuses for all jobs from the flow"""
 
@@ -197,9 +214,9 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                         when_humanized = humanize.naturaldate(info.when.astimezone())
                 rows.append(
                     [
-                        info.id,
-                        format_job_status(info.status),
-                        info.raw_id or "N/A",
+                        fmt_id(info.id),
+                        fmt_status(info.status),
+                        fmt_raw_id(info.raw_id or "N/A"),
                         when_humanized,
                     ]
                 )
@@ -213,47 +230,75 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             async for descr in self._resolve_jobs(meta_ctx, suffix):
                 await self._run_subproc("neuro", "status", descr.id)
         except ResourceNotFound:
-            click.echo(f"Job {click.style(job_id, bold=True)} is not running")
+            click.echo(f"Job {fmt_id(job_id)} is not running")
             sys.exit(1)
 
+    async def _try_attach_to_running(
+        self, job_id: str, suffix: Optional[str], args: Optional[Tuple[str]]
+    ) -> bool:
+        is_multi = await self.ctx.is_multi(job_id)
+
+        if not is_multi or suffix:
+            meta_ctx = await self._ensure_meta(job_id, suffix)
+            try:
+                jobs = []
+                async for descr in self._resolve_jobs(meta_ctx, suffix):
+                    jobs.append(descr)
+                if len(jobs) > 1:
+                    # Should never happen, but just in case
+                    raise click.ClickException(
+                        f"Found multiple running jobs for id {job_id} and"
+                        f" suffix {suffix}:\n"
+                        "\n".join(job.id for job in jobs)
+                    )
+                assert len(jobs) == 1
+                descr = jobs[0]
+                if is_multi and args:
+                    raise click.ClickException(
+                        "Multi job with such suffix is already running."
+                    )
+                if descr.status == JobStatus.PENDING:
+                    click.echo(f"Job {fmt_id(job_id)} is pending, " "try again later")
+                    sys.exit(2)
+                if descr.status == JobStatus.RUNNING:
+                    click.echo(f"Job {fmt_id(job_id)} is running, " "connecting...")
+                    if not is_multi:
+                        job_ctx = await meta_ctx.with_job(job_id)
+                        job = job_ctx.job
+                    else:
+                        assert suffix is not None
+                        multi_ctx = await meta_ctx.with_multi(suffix=suffix, args=())
+                        job_ctx = await multi_ctx.with_job(job_id)
+                        job = job_ctx.job
+                    # attach to job if needed, browse first
+                    if job.browse:
+                        await self._run_subproc("neuro", "job", "browse", descr.id)
+                    if not job.detach:
+                        await self._run_subproc("neuro", "attach", descr.id)
+                    return True
+                # Here the status is SUCCEED, CANCELLED or FAILED, restart
+            except ResourceNotFound:
+                # Job does not exist, run it
+                pass
+        return False
+
     async def run(
         self, job_id: str, suffix: Optional[str], args: Optional[Tuple[str]]
     ) -> None:
         """Run a named job"""
 
-        if not self.ctx.is_multi(job_id):
+        is_multi = await self.ctx.is_multi(job_id)
+
+        if not is_multi and args:
+            raise click.BadArgumentUsage(
+                "Additional job arguments are supported by multi-jobs only"
+            )
+
+        if await self._try_attach_to_running(job_id, suffix, args):
+            return  # Attached to running job
+
+        if not is_multi:
             meta_ctx = await self._ensure_meta(job_id, suffix)
-            if args:
-                raise click.BadArgumentUsage(
-                    "Additional job arguments are supported by multi-jobs only"
-                )
-            try:
-                async for descr in self._resolve_jobs(meta_ctx, suffix):
-                    # There is the only job for non-multi mode
-                    pass
-                if descr.status == JobStatus.PENDING:
-                    click.echo(
-                        f"Job {click.style(job_id, bold=True)} is pending, "
-                        "try again later"
-                    )
-                    sys.exit(2)
-                if descr.status == JobStatus.RUNNING:
-                    click.echo(
-                        f"Job {click.style(job_id, bold=True)} is running, "
-                        "connecting..."
-                    )
-                    job_ctx = await meta_ctx.with_job(job_id)
-                    job = job_ctx.job
-                    # attach to job if needed, browse first
-                    if job.browse:
-                        await self._run_subproc("neuro", "job", "browse", descr.id)
-                    if not job.detach:
-                        await self._run_subproc("neuro", "attach", descr.id)
-                    return
-                # Here the status is SUCCEDED or FAILED, restart
-            except ResourceNotFound:
-                # Job does not exist, run it
-                pass
             job_ctx = await meta_ctx.with_job(job_id)
             job = job_ctx.job
         else:
@@ -263,10 +308,6 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             multi_ctx = await meta_ctx.with_multi(suffix=suffix, args=args)
             job_ctx = await multi_ctx.with_job(job_id)
             job = job_ctx.job
-            if not job.detach:
-                raise RuntimeError(
-                    f"Multi-job {job_id} should use `detach: true` in config file"
-                )
 
         run_args = ["run"]
         if job.title:
@@ -290,7 +331,7 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             run_args.append(f"--env={k}={v}")
         for v in job.volumes:
             run_args.append(f"--volume={v}")
-        for t in job.tags:
+        for t in job_ctx.tags:
             run_args.append(f"--tag={t}")
         if job.life_span is not None:
             run_args.append(f"--life-span={int(job.life_span)}s")
@@ -317,7 +358,7 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             async for descr in self._resolve_jobs(meta_ctx, suffix):
                 await self._run_subproc("neuro", "logs", descr.id)
         except ResourceNotFound:
-            click.echo(f"Job {click.style(job_id, bold=True)} is not running")
+            click.echo(f"Job {fmt_id(job_id)} is not running")
             sys.exit(1)
 
     async def kill_job(self, job_id: str, suffix: Optional[str]) -> bool:
@@ -326,10 +367,10 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
 
         try:
             async for descr in self._resolve_jobs(meta_ctx, suffix):
-                if descr.status in (JobStatus.PENDING, JobStatus.RUNNING):
+                if descr.status in RUNNING_JOB_STATUSES:
                     await self.client.jobs.kill(descr.id)
                     descr = await self.client.jobs.status(descr.id)
-                    while descr.status not in (JobStatus.SUCCEEDED, JobStatus.FAILED):
+                    while descr.status not in TERMINATED_JOB_STATUSES:
                         await asyncio.sleep(0.2)
                         descr = await self.client.jobs.status(descr.id)
                     return True
@@ -340,9 +381,9 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
     async def kill(self, job_id: str, suffix: Optional[str]) -> None:
         """Kill named job"""
         if await self.kill_job(job_id, suffix):
-            click.echo(f"Killed job {click.style(job_id, bold=True)}")
+            click.echo(f"Killed job {fmt_id(job_id)}")
         else:
-            click.echo(f"Job {click.style(job_id, bold=True)} is not running")
+            click.echo(f"Job {fmt_id(job_id)} is not running")
 
     async def kill_all(self) -> None:
         """Kill all jobs"""
@@ -360,7 +401,7 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             await self.client.jobs.kill(descr.id)
             try:
                 descr = await self.client.jobs.status(descr.id)
-                while descr.status not in (JobStatus.SUCCEEDED, JobStatus.FAILED):
+                while descr.status not in TERMINATED_JOB_STATUSES:
                     await asyncio.sleep(0.2)
                     descr = await self.client.jobs.status(descr.id)
             except ResourceNotFound:
@@ -371,12 +412,12 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                 return job_id
 
         async for descr in self.client.jobs.list(
-            tags=self.ctx.defaults.tags, statuses=(JobStatus.PENDING, JobStatus.RUNNING)
+            tags=self.ctx.tags, statuses=RUNNING_JOB_STATUSES
         ):
             tasks.append(loop.create_task(kill(descr)))
 
         for job_id, ret in await asyncio.gather(*tasks):
-            click.echo(f"Killed job {click.style(job_id, bold=True)}")
+            click.echo(f"Killed job {fmt_id(job_id)}")
 
     # volumes subsystem
 
@@ -440,7 +481,7 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         for volume in self.ctx.volumes.values():
             if volume.local is not None:
                 volume_ctx = await self.find_volume(volume.id)
-                click.echo(f"Create volume {click.style(volume.id, bold=True)}")
+                click.echo(f"Create volume {fmt_id(volume.id)}")
                 await self._run_subproc(
                     "neuro",
                     "mkdir",
@@ -479,6 +520,10 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         cmd.append(f"--file={rel_dockerfile_path}")
         for arg in image_ctx.build_args:
             cmd.append(f"--build-arg={arg}")
+        for vol in image_ctx.volumes:
+            cmd.append(f"--volume={vol}")
+        for k, v in image_ctx.env.items():
+            cmd.append(f"--env={k}={v}")
         cmd.append(str(image_ctx.full_context_path))
         cmd.append(str(image_ctx.ref))
         await self._run_subproc("neuro-extras", "image", "build", *cmd)
