@@ -53,12 +53,8 @@ if sys.version_info < (3, 7):
 
 log = logging.getLogger(__name__)
 
-STARTED_RE: Final = re.compile(
-    r"\A\d+\.(?P<id>[a-zA-Z][a-zA-Z0-9_\-\.]*).started.json\Z"
-)
-FINISHED_RE: Final = re.compile(
-    r"\A\d+\.(?P<id>[a-zA-Z][a-zA-Z0-9_\-\.]*).finished.json\Z"
-)
+STARTED_RE: Final = re.compile(r"\A(?P<id>[a-zA-Z][a-zA-Z0-9_\-\.]*).started.json\Z")
+FINISHED_RE: Final = re.compile(r"\A(?P<id>[a-zA-Z][a-zA-Z0-9_\-\.]*).finished.json\Z")
 DIGITS = 4
 
 
@@ -115,6 +111,7 @@ class FinishedTask:
     finish_reason: str
     finish_description: str
     outputs: Mapping[str, str]
+    state: Mapping[str, str]
 
 
 # A storage abstraction
@@ -197,60 +194,153 @@ class BatchStorage(abc.ABC):
     async def finish_attempt(self, attempt: Attempt, result: JobStatus) -> None:
         pass
 
-    @abc.abstractmethod
     async def start_task(
         self,
         attempt: Attempt,
-        task_no: int,
         task_id: FullID,
         descr: JobDescription,
     ) -> StartedTask:
-        pass
+        assert descr.history.created_at is not None
+        ret = StartedTask(
+            attempt=attempt,
+            id=task_id,
+            raw_id=descr.id,
+            when=datetime.datetime.now(datetime.timezone.utc),
+            created_at=descr.history.created_at,
+        )
 
-    @abc.abstractmethod
-    async def start_batch_action(
+        await self.write_start(ret)
+        return ret
+
+    async def start_action(
         self,
         attempt: Attempt,
-        task_no: int,
         task_id: FullID,
     ) -> StartedTask:
-        pass
+        now = datetime.datetime.now(datetime.timezone.utc)
+        ret = StartedTask(
+            attempt=attempt,
+            id=task_id,
+            raw_id="",
+            when=now,
+            created_at=now,
+        )
+        await self.write_start(ret)
+        return ret
 
-    @abc.abstractmethod
     async def finish_task(
         self,
         attempt: Attempt,
-        task_no: int,
         task: StartedTask,
         descr: JobDescription,
         outputs: Mapping[str, str],
+        state: Mapping[str, str],
     ) -> FinishedTask:
-        pass
+        assert task.raw_id == descr.id
+        assert task.created_at == descr.history.created_at
+        assert descr.history.created_at is not None
+        assert descr.history.started_at is not None
+        assert descr.history.finished_at is not None
+        ret = FinishedTask(
+            attempt=attempt,
+            id=task.id,
+            raw_id=task.raw_id,
+            when=datetime.datetime.now(datetime.timezone.utc),
+            status=TaskStatus(descr.history.status),
+            exit_code=descr.history.exit_code,
+            created_at=descr.history.created_at,
+            started_at=descr.history.started_at,
+            finished_at=descr.history.finished_at,
+            finish_reason=descr.history.reason,
+            finish_description=descr.history.description,
+            outputs=outputs,
+            state=state,
+        )
+        await self.write_finish(ret)
+        return ret
 
-    @abc.abstractmethod
-    async def finish_batch_action(
+    async def finish_action(
         self,
         attempt: Attempt,
-        task_no: int,
         task: StartedTask,
         result: DepCtx,
     ) -> FinishedTask:
-        pass
+        now = datetime.datetime.now(datetime.timezone.utc)
+        assert result.result != TaskStatus.SKIPPED, (
+            "Finished task cannot be disabled (it is already started),"
+            " use .skip_task() instead"
+        )
+        status = TaskStatus(result.result)
+        ret = FinishedTask(
+            attempt=attempt,
+            id=task.id,
+            raw_id="",
+            when=now,
+            status=status,
+            exit_code=None,
+            created_at=task.created_at,
+            started_at=task.created_at,
+            finished_at=now,
+            finish_reason="",
+            finish_description="",
+            outputs=result.outputs,
+            state={},
+        )
+        await self.write_finish(ret)
+        return ret
 
-    @abc.abstractmethod
     async def skip_task(
         self,
         attempt: Attempt,
-        task_no: int,
         task_id: FullID,
-    ) -> FinishedTask:
+    ) -> Tuple[StartedTask, FinishedTask]:
+        now = datetime.datetime.now(datetime.timezone.utc)
+
+        st = StartedTask(
+            attempt=attempt,
+            id=task_id,
+            raw_id="",
+            when=now,
+            created_at=now,
+        )
+        await self.write_start(st)
+
+        ft = FinishedTask(
+            attempt=attempt,
+            id=task_id,
+            raw_id="",
+            when=now,
+            status=TaskStatus.SKIPPED,
+            exit_code=None,
+            created_at=now,
+            started_at=now,
+            finished_at=now,
+            finish_reason="",
+            finish_description="",
+            outputs={},
+            state={},
+        )
+        await self.write_finish(ft)
+        return st, ft
+
+    @abc.abstractmethod
+    async def write_start(
+        self,
+        task: StartedTask,
+    ) -> None:
+        pass
+
+    @abc.abstractmethod
+    async def write_finish(
+        self,
+        task: FinishedTask,
+    ) -> None:
         pass
 
     @abc.abstractmethod
     async def check_cache(
         self,
         attempt: Attempt,
-        task_no: int,
         task_id: FullID,
         ctx: TaskContext,
     ) -> Optional[FinishedTask]:
@@ -569,6 +659,7 @@ class BatchFSStorage(BatchStorage):
                 data = await self._read_json(attempt_url / fname)
                 assert match.group("id") == data["id"]
                 full_id = _id_from_json(data["id"])
+
                 started[full_id] = StartedTask(
                     attempt=attempt,
                     id=full_id,
@@ -595,6 +686,7 @@ class BatchFSStorage(BatchStorage):
                     finish_reason=data["finish_reason"],
                     finish_description=data["finish_description"],
                     outputs=data["outputs"],
+                    state=data["state"],
                 )
                 continue
             raise ValueError(f"Unexpected name {attempt_url / fname}")
@@ -608,186 +700,40 @@ class BatchFSStorage(BatchStorage):
         data = {"result": result.value}
         await self._write_json(attempt_url / f"{pre}.result.json", data)
 
-    async def start_task(
-        self,
-        attempt: Attempt,
-        task_no: int,
-        task_id: FullID,
-        descr: JobDescription,
-    ) -> StartedTask:
-        assert 0 <= task_no < int("9" * DIGITS) - 1, task_no
-        bake_uri = _mk_bake_uri(self._fs, attempt.bake)
-        attempt_url = bake_uri / f"{attempt.number:02d}.attempt"
-        pre = str(task_no + 1).zfill(DIGITS)
-        assert descr.history.created_at is not None
-        ret = StartedTask(
-            attempt=attempt,
-            id=task_id,
-            raw_id=descr.id,
-            when=datetime.datetime.now(datetime.timezone.utc),
-            created_at=descr.history.created_at,
-        )
+    async def write_start(self, st: StartedTask) -> None:
+        bake_uri = _mk_bake_uri(self._fs, st.attempt.bake)
+        attempt_url = bake_uri / f"{st.attempt.number:02d}.attempt"
 
         data = {
-            "id": _id_to_json(ret.id),
-            "raw_id": ret.raw_id,
-            "when": ret.when.isoformat(timespec="seconds"),
-            "created_at": ret.created_at.isoformat(timespec="seconds"),
+            "id": _id_to_json(st.id),
+            "raw_id": st.raw_id,
+            "when": st.when.isoformat(),
+            "created_at": st.created_at.isoformat(),
         }
-        await self._write_json(attempt_url / f"{pre}.{data['id']}.started.json", data)
-        return ret
+        await self._write_json(attempt_url / f"{data['id']}.started.json", data)
 
-    async def start_batch_action(
-        self,
-        attempt: Attempt,
-        task_no: int,
-        task_id: FullID,
-    ) -> StartedTask:
-        assert 0 <= task_no < int("9" * DIGITS) - 1, task_no
-        bake_uri = _mk_bake_uri(self._fs, attempt.bake)
-        attempt_url = bake_uri / f"{attempt.number:02d}.attempt"
-        pre = str(task_no + 1).zfill(DIGITS)
-        now = datetime.datetime.now(datetime.timezone.utc)
-        ret = StartedTask(
-            attempt=attempt,
-            id=task_id,
-            raw_id="",
-            when=now,
-            created_at=now,
-        )
-
-        data = {
-            "id": _id_to_json(ret.id),
-            "raw_id": ret.raw_id,
-            "when": ret.when.isoformat(timespec="seconds"),
-            "created_at": ret.created_at.isoformat(timespec="seconds"),
-        }
-        await self._write_json(attempt_url / f"{pre}.{data['id']}.started.json", data)
-        return ret
-
-    async def finish_task(
-        self,
-        attempt: Attempt,
-        task_no: int,
-        task: StartedTask,
-        descr: JobDescription,
-        outputs: Mapping[str, str],
-    ) -> FinishedTask:
-        assert task.raw_id == descr.id
-        assert task.created_at == descr.history.created_at
-        assert descr.history.created_at is not None
-        assert descr.history.started_at is not None
-        assert descr.history.finished_at is not None
-        ret = FinishedTask(
-            attempt=attempt,
-            id=task.id,
-            raw_id=task.raw_id,
-            when=datetime.datetime.now(datetime.timezone.utc),
-            status=TaskStatus(descr.history.status),
-            exit_code=descr.history.exit_code,
-            created_at=descr.history.created_at,
-            started_at=descr.history.started_at,
-            finished_at=descr.history.finished_at,
-            finish_reason=descr.history.reason,
-            finish_description=descr.history.description,
-            outputs=outputs,
-        )
-        await self._write_finish(attempt, task_no, ret)
-        return ret
-
-    async def _write_finish(
-        self, attempt: Attempt, task_no: int, ft: FinishedTask
-    ) -> None:
-        assert 0 <= task_no < int("9" * DIGITS) - 1, task_no
-        bake_uri = _mk_bake_uri(self._fs, attempt.bake)
-        attempt_url = bake_uri / f"{attempt.number:02d}.attempt"
-        pre = str(task_no + 1).zfill(DIGITS)
+    async def write_finish(self, ft: FinishedTask) -> None:
+        bake_uri = _mk_bake_uri(self._fs, ft.attempt.bake)
+        attempt_url = bake_uri / f"{ft.attempt.number:02d}.attempt"
         data = {
             "id": _id_to_json(ft.id),
             "raw_id": ft.raw_id,
-            "when": ft.when.isoformat(timespec="seconds"),
+            "when": ft.when.isoformat(),
             "status": ft.status.value,
             "exit_code": ft.exit_code,
-            "created_at": ft.created_at.isoformat(timespec="seconds"),
-            "started_at": ft.started_at.isoformat(timespec="seconds"),
-            "finished_at": ft.finished_at.isoformat(timespec="seconds"),
+            "created_at": ft.created_at.isoformat(),
+            "started_at": ft.started_at.isoformat(),
+            "finished_at": ft.finished_at.isoformat(),
             "finish_reason": ft.finish_reason,
             "finish_description": ft.finish_description,
             "outputs": ft.outputs,
+            "state": ft.state,
         }
-        await self._write_json(attempt_url / f"{pre}.{data['id']}.finished.json", data)
-
-    async def finish_batch_action(
-        self,
-        attempt: Attempt,
-        task_no: int,
-        task: StartedTask,
-        result: DepCtx,
-    ) -> FinishedTask:
-        now = datetime.datetime.now(datetime.timezone.utc)
-        assert result.result != TaskStatus.SKIPPED, (
-            "Finished task cannot be disabled (it is already started),"
-            " use .skip_task() instead"
-        )
-        status = TaskStatus(result.result)
-        ret = FinishedTask(
-            attempt=attempt,
-            id=task.id,
-            raw_id="",
-            when=now,
-            status=status,
-            exit_code=None,
-            created_at=task.created_at,
-            started_at=task.created_at,
-            finished_at=now,
-            finish_reason="",
-            finish_description="",
-            outputs=result.outputs,
-        )
-        await self._write_finish(attempt, task_no, ret)
-        return ret
-
-    async def skip_task(
-        self,
-        attempt: Attempt,
-        task_no: int,
-        task_id: FullID,
-    ) -> FinishedTask:
-        assert 0 <= task_no < int("9" * DIGITS) - 1, task_no
-        bake_uri = _mk_bake_uri(self._fs, attempt.bake)
-        attempt_url = bake_uri / f"{attempt.number:02d}.attempt"
-        pre = str(task_no + 1).zfill(DIGITS)
-        now = datetime.datetime.now(datetime.timezone.utc)
-
-        data = {
-            "id": _id_to_json(task_id),
-            "raw_id": "",
-            "when": now.isoformat(timespec="seconds"),
-            "created_at": now.isoformat(timespec="seconds"),
-        }
-        await self._write_json(attempt_url / f"{pre}.{data['id']}.started.json", data)
-
-        ret = FinishedTask(
-            attempt=attempt,
-            id=task_id,
-            raw_id="",
-            when=now,
-            status=TaskStatus.SKIPPED,
-            exit_code=None,
-            created_at=now,
-            started_at=now,
-            finished_at=now,
-            finish_reason="",
-            finish_description="",
-            outputs={},
-        )
-        await self._write_finish(attempt, task_no, ret)
-        return ret
+        await self._write_json(attempt_url / f"{data['id']}.finished.json", data)
 
     async def check_cache(
         self,
         attempt: Attempt,
-        task_no: int,
         task_id: FullID,
         ctx: TaskContext,
     ) -> Optional[FinishedTask]:
@@ -826,8 +772,9 @@ class BatchFSStorage(BatchStorage):
                 finish_reason=data["finish_reason"],
                 finish_description=data["finish_description"],
                 outputs=data["outputs"],
+                state=data["state"],
             )
-            await self._write_finish(attempt, task_no, ret)
+            await self.write_finish(ret)
             return ret
         except (KeyError, ValueError, TypeError):
             # something is wrong with stored JSON,
@@ -851,17 +798,18 @@ class BatchFSStorage(BatchStorage):
         assert ft.raw_id is not None, (ft.id, ft.raw_id)
 
         data = {
-            "when": ft.when.isoformat(timespec="seconds"),
+            "when": ft.when.isoformat(),
             "digest": ctx_digest,
             "raw_id": ft.raw_id,
             "status": ft.status.value,
             "exit_code": ft.exit_code,
-            "created_at": ft.created_at.isoformat(timespec="seconds"),
-            "started_at": ft.started_at.isoformat(timespec="seconds"),
-            "finished_at": ft.finished_at.isoformat(timespec="seconds"),
+            "created_at": ft.created_at.isoformat(),
+            "started_at": ft.started_at.isoformat(),
+            "finished_at": ft.finished_at.isoformat(),
             "finish_reason": ft.finish_reason,
             "finish_description": ft.finish_description,
             "outputs": ft.outputs,
+            "state": ft.state,
         }
         try:
             await self._write_json(url, data, overwrite=True)
@@ -903,7 +851,7 @@ class BatchFSStorage(BatchStorage):
         self, url: URL, data: Dict[str, Any], *, overwrite: bool = False
     ) -> None:
         if not data.get("when"):
-            data["when"] = _dt2str(_now())
+            data["when"] = _now().isoformat()
 
         await self._write_file(url, json.dumps(data), overwrite=overwrite)
 
@@ -937,7 +885,7 @@ def _bake_to_json(bake: Bake) -> Dict[str, Any]:
     return {
         "project": bake.project,
         "batch": bake.batch,
-        "when": _dt2str(bake.when),
+        "when": bake.when.isoformat(),
         "suffix": bake.suffix,
         "graphs": graphs,
     }
@@ -974,7 +922,7 @@ def _attempt_to_json(attempt: Attempt) -> Dict[str, Any]:
     return {
         "number": attempt.number,
         "bake": _bake_to_json(attempt.bake),
-        "when": _dt2str(attempt.when),
+        "when": attempt.when.isoformat(),
     }
 
 
